package moe.nekojimi.nodeprocessor;

import java.io.*;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.time.OffsetDateTime;
import java.util.*;
import java.util.stream.Collectors;
import moe.nekojimi.nodeprocessor.links.Link;
import moe.nekojimi.nodeprocessor.nodes.MethodNode;
import moe.nekojimi.nodeprocessor.nodes.DebugOutNode;
import moe.nekojimi.nodeprocessor.nodes.ValueNode;
import org.reflections.Reflections;
import org.reflections.scanners.SubTypesScanner;

/**
 *
 * @author Nekojimi
 */
public class Main
{

    private static final Class[] testClasses =
    {
        String.class,
        Integer.class,
        Math.class,
        HashMap.class,
        OffsetDateTime.class,

    };

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws NoSuchMethodException, ClassNotFoundException, IOException
    {
        for (Class testClass : testClasses)
        {
            System.out.println(testClass.getName() + ":");
            Method[] methods = testClass.getMethods();
            for (Method method : methods)
            {
                if (method.getDeclaringClass() == testClass)
                {
                    ClassMethodType type = guessMethodType(method);
                    System.out.println("\t" + method.toString());
                    System.out.println("\t\t" + type.name());
                }
            }
        }
//        helloWorldProgram();
    }

    private static void fibonnaciProgram()
    {
        Program program = new Program();
    }

    private static void addNumbersProgram() throws NoSuchMethodException
    {
        Program program = new Program();
        ValueNode<Integer> twoNode = new ValueNode<>(program, Integer.class);
        twoNode.setValue(2);
        ValueNode<Integer> threeNode = new ValueNode<>(program, Integer.class);
        threeNode.setValue(3);

        MethodNode<Integer> methodNode = new MethodNode<>(program, Integer.class, Integer.class.getMethod("sum", Integer.class, Integer.class));

        DebugOutNode debugOutNode = new DebugOutNode(program);

//        Link.makeLink(Integer.class, twoNode, fromPort, twoNode, toPort);

        program.start();
    }

    private static void helloWorldProgram() throws SecurityException, NoSuchMethodException
    {
        Program program = new Program();
        ValueNode<String> valueNode = new ValueNode<>(program, String.class);
        valueNode.setValue("Hello world!");
        final Method method = String.class.getMethod("toUpperCase");
        MethodNode<String> methodNode = new MethodNode<>(program, String.class, method);
        DebugOutNode debugOutNode = new DebugOutNode(program);

        Link.makeLink(String.class, valueNode, "out", methodNode, "object");
        Link.makeLink(String.class, methodNode, "out", debugOutNode, "in");

        program.start();
    }

    public static Set<Class> findAllClassesUsingClassLoader(String packageName)
    {
        InputStream stream = ClassLoader.getSystemClassLoader()
                .getResourceAsStream(packageName.replaceAll("[.]", "/"));
        BufferedReader reader = new BufferedReader(new InputStreamReader(stream));
        return reader.lines()
                .filter(line -> line.endsWith(".class"))
                .map(line -> getClass(line, packageName))
                .collect(Collectors.toSet());
    }

    public static Set<Class> findAllClassesUsingReflectionsLibrary(String packageName)
    {
        Reflections reflections = new Reflections(packageName, new SubTypesScanner(false));
        return reflections.getSubTypesOf(Object.class)
                .stream()
                .collect(Collectors.toSet());
    }

    private static Class getClass(String className, String packageName)
    {
        try
        {
            return Class.forName(packageName + "."
                    + className.substring(0, className.lastIndexOf('.')));
        } catch (ClassNotFoundException e)
        {
            // handle the exception
        }
        return null;
    }

    private static ClassMethodType guessMethodType(Method method)
    {
        String name = method.getName();
        Class<?> clazz = method.getDeclaringClass();
        Class<?> returnType = method.getReturnType();
        boolean hasReturn = (returnType != Void.TYPE);
        int paramCount = method.getParameterCount();
        Class<?> firstParamType = null;
        Parameter[] parameters = method.getParameters();
        if (paramCount > 0)
            firstParamType = parameters[0].getType();
        boolean isStatic = Modifier.isStatic(method.getModifiers());

        if (!isStatic)
        {
            if (hasReturn)
            {
                if (paramCount == 0) // methods that only return
                {
                    if (name.startsWith("take") || name.startsWith("poll") || name.startsWith("remove") || name.startsWith("pop") || name.startsWith("read"))
                        return ClassMethodType.TAKER;
                    else
                        return ClassMethodType.GETTER;
                } else if (clazz.isAssignableFrom(returnType))
                    return ClassMethodType.TRANSFORMER;
                else if (paramCount == 1
                        && Modifier.isAbstract(parameters[0].getType().getModifiers())
                        && !parameters[0].getType().isPrimitive()
                        && name.toLowerCase().matches("(on.+|.*listener.*|.*observer.*)"))
                {
                    int abstractMethods = 0;
                    Method[] paramTypeMethods = parameters[0].getType().getMethods();
                    for (Method paramTypeMethod : paramTypeMethods)
                    {
                        if (Modifier.isAbstract(paramTypeMethod.getModifiers()))
                            abstractMethods++;
                    }
                    if (abstractMethods >= 1)
                        return ClassMethodType.CALLBACK;
                } else
                    return ClassMethodType.FUNCTION;
            } else // returns nothing
            {
                if (paramCount == 1)
                {
                    if (name.startsWith("give") || name.startsWith("add") || name.startsWith("push") || name.startsWith("write"))
                        return ClassMethodType.ADDER;
                    else
                        return ClassMethodType.SETTER;
                } else if (paramCount == 0) // no  input, return nothing
                    return ClassMethodType.ACTION;
            }
        } else // static method
        {
            if (clazz.isAssignableFrom(returnType))
                return ClassMethodType.FACTORY;
            else if (hasReturn && paramCount > 0)
                return ClassMethodType.FUNCTION;
            else if (hasReturn && paramCount == 0)
                return ClassMethodType.CONSTANT;
        }

        return ClassMethodType.OTHER;
    }

}
