/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.nodeprocessor.nodes;

import java.io.Closeable;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Parameter;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import moe.nekojimi.nodeprocessor.Program;
import moe.nekojimi.nodeprocessor.ports.InPort;
import moe.nekojimi.nodeprocessor.ports.OutPort;

/**
 * Represents an object that performs some function.
 * Has a set of inputs that will cause the object to be (re) created when
 * changed.
 *
 * @author jimj316
 * @param <T>
 */
public class ObjectNode<T> extends Node
{
    protected T object;
    private final Constructor<T> constructor;
    private final Set<InPort> composingInPorts = new HashSet<>();
    private final Set<InPort> setterInPorts = new HashSet<>();
    private final Set<OutPort> getterOutPorts = new HashSet<>();
    private final OutPort<T> objectOutPort;
    private final Map<String, Object> composingValues = new HashMap<>();
    private final Class<T> clazz;

    public ObjectNode(Program scene, Class<T> clazz, Constructor<T> constructor)
    {
        super(scene);
        this.clazz = clazz;
        this.constructor = constructor;
        objectOutPort = new OutPort<>("this", clazz, this);

        addConstructor();
        addAutoMethods();
    }

    public void addAutoMethods() throws SecurityException
    {
        for (Method method : clazz.getMethods())
        {
            String name = method.getName();
            Class<?> returnType = method.getReturnType();
            Parameter[] parameters = method.getParameters();
            if (returnType == null && parameters.length == 1 && name.startsWith("set"))
            {
                // this is a setter method
                InPort in = new InPort(name, parameters[0].getType(), this);
                addInPort(in);
                setterInPorts.add(in);
            } else if (returnType != null && parameters.length == 0 && (name.startsWith("get") || name.startsWith("is")))
            {
                // this is a getter method
                OutPort out = new OutPort(name, returnType, this);
                addOutPort(out);
                getterOutPorts.add(out);
            }
        }
    }

    public void addConstructor()
    {
        Parameter[] parameters = constructor.getParameters();
        for (Parameter parameter : parameters)
        {
            InPort in = new InPort(parameter.getName(), parameter.getType(), this);
            addInPort(in);
            composingInPorts.add(in);
        }
    }

    @Override
    public void reset()
    {
        super.reset();
        if (object instanceof Closeable)
        {
            try
            {
                ((Closeable) object).close();
            } catch (IOException ex)
            {
                Logger.getLogger(ObjectNode.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        object = null;
    }

    @Override
    public boolean process()
    {
        // check if the composing values have changed; if so delete the object

        // if the object doesn't exist, create it
        if (object == null)
        {
            Constructor<?>[] constructors = clazz.getConstructors();
            for (Constructor constructor : constructors)
            {
                Parameter[] parameters = constructor.getParameters();
                List<Object> args = new ArrayList<>();
                for (Parameter parameter : parameters)
                {
                    InPort<?> inPort = getInPort(parameter.getName(), parameter.getType());
                    if (inPort.has())
                        args.add(inPort.get());
                    else
                        break;
                }
                if (args.size() == parameters.length)
                {
                    try
                    {
                        object = (T) constructor.newInstance(args.toArray());
                        objectOutPort.set(object);
                    } catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex)
                    {
                        Logger.getLogger(ObjectNode.class.getName()).log(Level.SEVERE, null, ex);
                    }
                    break;
                }
            }
        }

        // now the object exists, do something with it
        if (object != null)
        {
            // take input from setters
            // give output to getters

            return doProcess();
        }
        return true;
    }

    protected boolean doProcess()
    {
        return true;
    }

    @Override
    public boolean check()
    {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
}
